@@ -1,3 +1,4 @@
+#define PY_SSIZE_T_CLEAN
 #include "Python.h"
 #if PY_MAJOR_VERSION >= 3
 #	define isPy3
@@ -20,7 +21,7 @@
 #endif
 
 
-#define VERSION "3.00"
+#define VERSION "4.00"
 #define MODULENAME "_renderPM"
 #ifdef isPy3
 #	define PyInt_FromLong	PyLong_FromLong
@@ -42,35 +43,45 @@ PyObject *RLPy_FindMethod(PyMethodDef *ml, PyObject *self, const char* name){
 #		define Py_TYPE(ob) (((PyObject*)(ob))->ob_type)
 #	endif
 #endif
-
+#define __STR(x) #x
+#define STRINGIFY(x) __STR(x)
 #ifndef LIBART_VERSION
-#	define LIBART_VERSION "?.?.?"
+#	define LIBART_VERSION ?.?.?
 #endif
 #ifdef	RENDERPM_FT
-#	define _FT_DOC "    ft_get_face(fontName) --> ft_face instance\n"
+#	define _FT_DOC "    _renderPM.ft_get_face(fontName) --> ft_face instance\n"
 #else
 #	define _FT_DOC ""
 #endif
+#ifdef MEMORY_DEBUG
+#	define _MDBG_DOC "    _renderPM.mtrace(int) start or stop the malloc tracing\n"
+#else
+#	define _MDBG_DOC ""
+#endif
 PyDoc_STRVAR(__DOC__,
 "Helper extension module for renderPM.\n\
 \n\
 Interface summary:\n\
 \n\
-	import _renderPM\n\
-	gstate(width,height[,depth=3,bg=0xffffff]) #create an initialised graphics state\n\
-	makeT1Font(fontName,pfbPath,names[,reader])	#make a T1 font\n\
-	delCache() #delete all T1 font info\n\
-	pil2pict(cols,rows,datastr,palette) return PICT version of im as bytes\n"
+    from reportlan.graphics import _renderPM\n\
+    _renderPM.gstate(width,height[,depth=3,bg=0xffffff]) create an \n\
+        initialised graphics state\n\
+    _renderPM.makeT1Font(fontName,pfbPath,names[,reader]) make a T1 font\n\
+    _renderPM.delCache() delete all T1 font info\n\
+    _renderPM.pil2pict(cols,rows,datastr,palette) return PICT version of\n\
+        im as bytes\n"
 _FT_DOC
+_MDBG_DOC
 "\n\
-	_libart_version	# base library version string\n\
-	_version		# module version string\n\
+    _renderPM._libart_version base library version string\n\
+    _renderPM._version module version string ie " VERSION "\n\
 ");
 
 #if PY_VERSION_HEX < 0x01060000
 #	define PyObject_DEL(op) PyMem_DEL((op))
 #endif
 
+#define VECSP 0.25
 
 typedef struct {
   int format;
@@ -92,7 +103,8 @@ typedef	struct {
 static	PyObject* parse_utf8(PyObject* self, PyObject* args)
 {
 	char		*c, *msg;
-	int			n, i;
+	Py_ssize_t	n;
+	int			i;
 	unsigned	first, second, third;
 	PyObject	*r;
     if(!PyArg_ParseTuple(args, "t#:parse_utf8", &c, &n)) return NULL;
@@ -185,7 +197,7 @@ static py_FT_FontObject *_get_ft_face(char *fontName)
 	_data = PyObject_GetAttrString(face,"_ttf_data");
 	Py_DECREF(face);
 	if(!_data) goto RET;
-	error = FT_New_Memory_Face(ft_library, (unsigned char *)PyBytes_AsString(_data), PyBytes_GET_SIZE(_data), 0, &ft_face->face);
+	error = FT_New_Memory_Face(ft_library, (unsigned char *)PyBytes_AsString(_data), (FT_Long)PyBytes_GET_SIZE(_data), 0, &ft_face->face);
 	Py_DECREF(_data);
 	if(error){
 		PyErr_Format(PyExc_IOError, "FT_New_Memory_Face(%s) Failed!", fontName);
@@ -260,7 +272,7 @@ static pixBufT* pixBufAlloc(int w, int h, int nchan, gstateColorX bg)
 			p->width = w;
 			p->height = h;
 			p->nchan = nchan;
-			p->rowstride = stride;
+			p->rowstride = (int)stride;
 
 			/*set up the background*/
 			if(bg.stride==0){	/*simple color case*/
@@ -325,6 +337,7 @@ typedef struct {
 	double		strokeOpacity;
 	gstateColor	fillColor;			/*fill color*/
 	int			fillMode;
+	int			textRenderMode;
 	double		fillOpacity;
 	double		fontSize;
 	double		fontEMSize;			/*for scaling to points*/
@@ -605,23 +618,49 @@ static double _vpath_area(ArtVpath *p)
 	return a;
 }
 
-static	PyObject* gstate_clipPathSet(gstateObject* self, PyObject* args)
-{
+static	void _gstate_clipPathSetOrAdd(gstateObject* self, int fillMode, int add, int endIt){
 	ArtVpath	*vpath;
 	ArtVpath	*trVpath;
-	int			fillMode = self->fillMode;
-
-	if(!PyArg_ParseTuple(args,"|i:clipPathSet",&fillMode)) return NULL;
-	gstate_pathEnd(self);
+	if (fillMode==-1) fillMode = self->fillMode;
+	if(endIt) gstate_pathEnd(self);
 	dump_path(self);
-	vpath = art_bez_path_to_vec(self->path, 0.25);
+	vpath = art_bez_path_to_vec(self->path, VECSP);
 	dump_vpath("after -->vec",vpath);
 	trVpath = art_vpath_affine_transform (vpath, self->ctm);
 	_vpath_area(trVpath);
-	if(self->clipSVP) art_svp_free(self->clipSVP);
-	self->clipSVP = art_svp_from_vpath(trVpath);
+	if (add) {
+		ArtSVP		*newSVP = art_svp_from_vpath(trVpath);
+		if(self->clipSVP) {
+			ArtSVP *oldSVP = self->clipSVP;
+			self->clipSVP = art_svp_union(oldSVP,newSVP);
+			art_svp_free(oldSVP);
+			art_svp_free(newSVP);
+			}
+		else{
+			self->clipSVP = newSVP;
+			}
+		}
+	else{
+		if(self->clipSVP) art_svp_free(self->clipSVP);
+		self->clipSVP = art_svp_from_vpath(trVpath);
+		}
 	art_free(trVpath);
 	art_free(vpath);
+}
+
+static	PyObject* gstate_clipPathSet(gstateObject* self, PyObject* args)
+{
+	int			fillMode = -1;
+	if(!PyArg_ParseTuple(args,"|i:clipPathSet",&fillMode)) return NULL;
+	_gstate_clipPathSetOrAdd(self, fillMode, 0, 1);
+	Py_INCREF(Py_None);
+	return Py_None;
+}
+static	PyObject* gstate_clipPathAdd(gstateObject* self, PyObject* args)
+{
+	int			fillMode = -1;
+	if(!PyArg_ParseTuple(args,"|i:clipPathAdd",&fillMode)) return NULL;
+	_gstate_clipPathSetOrAdd(self, fillMode, 1, 1);
 	Py_INCREF(Py_None);
 	return Py_None;
 }
@@ -637,15 +676,15 @@ static void _gstate_pathFill(gstateObject* self,int endIt, int vpReverse, int fi
 		double		a;
 		if(endIt) gstate_pathEnd(self);
 		dump_path(self);
-		vpath = art_bez_path_to_vec(self->path, 0.25);
+		vpath = art_bez_path_to_vec(self->path, VECSP);
 		if(0 && vpReverse) _vpath_reverse(vpath);
 		trVpath =  art_vpath_affine_transform(vpath, self->ctm);
 		a = _vpath_area(trVpath);
 		if(fabs(a)>1e-7){
 			/*fill only larger things*/
 			tmp_vpath =  art_vpath_perturb(trVpath);
-			trVpath =  art_vpath_perturb(tmp_vpath);
-			art_free(tmp_vpath);
+			art_free(trVpath);	/*free the original*/
+			trVpath =  tmp_vpath;
 			svp = art_svp_from_vpath(trVpath);
 			dump_svp("fill svp from vpath",svp);
 			if(fillMode==0){
@@ -695,18 +734,17 @@ static PyObject* gstate_pathFill(gstateObject* self, PyObject* args)
 	return Py_None;
 }
 
-static PyObject* gstate_pathStroke(gstateObject* self, PyObject* args)
+static void _gstate_pathStroke(gstateObject* self, int endIt)
 {
 	ArtVpath	*vpath=NULL, *trVpath;
 	ArtSVP*		svp=NULL;
 	ArtSVP*		tmp_svp=NULL;
 	pixBufT*	p;
 
-	if(!PyArg_ParseTuple(args,":pathStroke")) return NULL;
 	if(self->strokeColor.valid && self->strokeWidth>0){
-		gstate_pathEnd(self);
+		if(endIt) gstate_pathEnd(self);
 		dump_path(self);
-		vpath = art_bez_path_to_vec(self->path, 0.25);
+		vpath = art_bez_path_to_vec(self->path, VECSP);
 
 		if(self->dash.dash){
 			ArtVpath*	tvpath=vpath;
@@ -738,6 +776,12 @@ static PyObject* gstate_pathStroke(gstateObject* self, PyObject* args)
 		art_svp_free(svp);
 		art_free(vpath);
 		}
+}
+
+static PyObject* gstate_pathStroke(gstateObject* self, PyObject* args)
+{
+	if(!PyArg_ParseTuple(args,":pathStroke")) return NULL;
+	_gstate_pathStroke(self, 1);
 	Py_INCREF(Py_None);
 	return Py_None;
 }
@@ -835,7 +879,8 @@ static PyObject* gstate_drawString(gstateObject* self, PyObject* args)
 	A2DMX	orig, trans = {1,0,0,1,0,0}, scaleMat = {1,0,0,1,0,0};
 	double	scaleFactor, x, y, w;
 	char*	text;
-	int		c, textlen, i;
+	int		c, i, textRenderMode=self->textRenderMode;
+	Py_ssize_t	textlen;
 	ArtBpath	*saved_path, *path;
 	void	*font = self->font;
 	PyObject	*textObj, *obj0;
@@ -928,7 +973,13 @@ static PyObject* gstate_drawString(gstateObject* self, PyObject* args)
 
 		if(path){
 			self->path = path;
-			_gstate_pathFill(self,0,1,1);
+			if (textRenderMode==0 || textRenderMode==2 || textRenderMode==4 || textRenderMode==6)
+				_gstate_pathFill(self,0,1,1);
+			if (textRenderMode==1 || textRenderMode==2 || textRenderMode==5 || textRenderMode==6)
+				_gstate_pathStroke(self,0);
+			if (textRenderMode>=4)
+				_gstate_clipPathSetOrAdd(self, -1, 1, 0);	/*we are adding*/
+
 #ifdef	RENDERPM_FT
 			if(!ft_font && path!=notdefPath)
 #endif
@@ -1021,7 +1072,7 @@ static PyObject* _get_gstateVPath(gstateObject *self)
 	int			i;
 
 	gstate_pathEnd(self);
-	v = vpath = art_bez_path_to_vec(self->path, 0.25);
+	v = vpath = art_bez_path_to_vec(self->path, VECSP);
 	while(v->code!=ART_END) v++;
 	P = PyTuple_New(v-vpath);
 	i = 0;
@@ -1055,7 +1106,8 @@ static PyObject* gstate__stringPath(gstateObject* self, PyObject* args)
 	char*	text;
 	PyObject *P, *p;
 	ArtBpath	*path, *pp;
-	int		textlen, i, c;
+	Py_ssize_t	textlen;
+	int		i, c;
 	void	*font = self->font;
 	PyObject	*textObj, *obj0;
 #ifdef	RENDERPM_FT
@@ -1140,7 +1192,7 @@ static PyObject* gstate__stringPath(gstateObject* self, PyObject* args)
 				pp->y3 = pp->y3*s+y;
 				pp++;
 				}
-			p = _get_gstatePath(pp-path,path);
+			p = _get_gstatePath((int)(pp-path),path);
 #ifdef	RENDERPM_FT
 			if(!ft_font && path!=notdefPath)
 #endif
@@ -1326,7 +1378,8 @@ static	int _set_gstateDashArray(PyObject* value, gstateObject* self)
 		return 1;
 		}
 	else {
-		int	n_dash, i, r=0;
+		Py_ssize_t	n_dash;
+		int i, r=0;
 		PyObject	*v=NULL, *pDash=NULL;
 		double		offset, *dash=NULL;
 		if(!PySequence_Check(value) || PySequence_Length(value)!=2){
@@ -1349,7 +1402,7 @@ L1:			_safeDecr(&v);
 
 		/*everything checks out release current thing and set new one*/
 		_dashFree(self);
-		self->dash.n_dash = n_dash;
+		self->dash.n_dash = (int)n_dash;
 		self->dash.offset = offset;
 		self->dash.dash = dash;
 		r = 1;
@@ -1428,7 +1481,7 @@ static	int _set_gstateColorX(PyObject* value, gstateColorX* c)
 {
 	int	i;
 	if(PySequence_Check(value)){
-		size_t	len;
+		Py_ssize_t	len;
 		i = PyArg_Parse(value,"(iis#)",&c->width,&c->height,&c->buf,&len);
 		if(i){
 			/*we assume depth 3*/
@@ -1496,6 +1549,7 @@ static PyObject* _get_gstateFontName(gstateObject *self)
 static struct PyMethodDef gstate_methods[] = {
 	{"clipPathClear", (PyCFunction)gstate_clipPathClear, METH_VARARGS, "clipPathClear()"},
 	{"clipPathSet", (PyCFunction)gstate_clipPathSet, METH_VARARGS, "clipPathSet([fillMode])"},
+	{"clipPathAdd", (PyCFunction)gstate_clipPathAdd, METH_VARARGS, "clipPathAdd([fillMode])"},
 	{"curveTo", (PyCFunction)gstate_curveTo, METH_VARARGS, "curveTo(x1,y1,x2,y2,x3,y3)"},
 	{"drawString", (PyCFunction)gstate_drawString, METH_VARARGS, "drawString(x,y,text)"},
 	{"lineTo", (PyCFunction)gstate_lineTo, METH_VARARGS, "lineTo(x,y)"},
@@ -1536,6 +1590,7 @@ static PyObject* gstate_getattr(gstateObject *self, char *name)
 	else if(!strcmp(name,"fontName")) return _get_gstateFontName(self);
 	else if(!strcmp(name,"fontNameI")) return _get_gstateFontNameI(self);
 	else if(!strcmp(name,"dashArray")) return _get_gstateDashArray(self);
+	else if(!strcmp(name,"textRenderMode")) return PyInt_FromLong(self->textRenderMode);
 	else if(!strcmp(name,"pixBuf")){
 		pixBufT* p = self->pixBuf;
 		int	nw = p->rowstride;
@@ -1574,6 +1629,7 @@ static int gstate_setattr(gstateObject *self, char *name, PyObject* value)
 	else if(!strcmp(name,"strokeOpacity")) i = PyArg_Parse(value,"d",&self->strokeOpacity);
 	else if(!strcmp(name,"fillOpacity")) i = PyArg_Parse(value,"d",&self->fillOpacity);
 	else if(!strcmp(name,"dashArray")) i = _set_gstateDashArray(value,self);
+	else if(!strcmp(name,"textRenderMode")) i = PyArg_Parse(value,"i",&self->textRenderMode);
 	else {
 		PyErr_SetString(PyExc_AttributeError, name);
 		i = 0;
@@ -1606,6 +1662,7 @@ static	void gstateFree(gstateObject* self)
 gstates have the following methods\n\
  clipPathClear() clear clipPath\n\
  clipPathSet([fillMode]) move current path into clipPath\n\
+ clipPathAdd([fillMode]) add current path into clipPath\n\
  curveTo(x1,y1,x2,y2,x3,y3)  #add a curveTo type segment\n\
  drawString(x,y,text)\n\
  moveTo(x,y) start a segment\n\
@@ -1626,6 +1683,7 @@ ctm			6vec float transformation matrix\n\
 strokeColor 32bit stroke colour\n\
 fillColor	32bit fill colour\n\
 fillMode	int fill rule\n\
+textRenderMode	int controls filling and stroking (follows pdf values)\n\
 lineCap		int\n\
 lineJoin	int\n\
 hasClipPath readonly int\n\
@@ -1725,6 +1783,7 @@ static	gstateObject* gstate(PyObject* module, PyObject* args, PyObject* keywds)
 			self->strokeColor.valid = self->fillColor.valid = 0;
 			self->lineCap = self->lineJoin = 0;
 			self->fillMode = FILL_NON_ZERO;	/*our default for renderPM differed from PDF/PS*/
+			self->textRenderMode = 0;
 			self->strokeOpacity = self->strokeWidth = self->fillOpacity = 1.0;
 			self->pathLen = 0;
 			self->pathMax = m;
@@ -1753,7 +1812,7 @@ static	char* my_pfb_reader(void *data, const char *filename, int *psize)
 		/*the file should have been read as binary*/
 		if(PyBytes_Check(result)){
 			char	*pystr = PyBytes_AS_STRING(result);
-			int		size = PyBytes_GET_SIZE(result);
+			int		size = (int)PyBytes_GET_SIZE(result);
 			*psize = size;
 			memcpy(pfb=gt1_alloc(size),pystr,size);
 			}
@@ -1821,7 +1880,7 @@ static	PyObject*	makeT1Font(PyObject* self, PyObject *args, PyObject *kw)
 			rfunc.data = reader;
 			rfunc.reader = my_pfb_reader;
 			}
-		if(!gt1_create_encoded_font(name,pfbPath,names,N,prfunc)){
+		if(!gt1_create_encoded_font(name,pfbPath,names,(int)N,prfunc)){
 			PyErr_SetString(PyExc_ValueError, "_renderPM.makeT1Font: can't make font");
 			ok = 0;
 			}
@@ -1843,6 +1902,22 @@ static PyObject* delCache(PyObject* self, PyObject* args)
 	Py_INCREF(Py_None);
 	return Py_None;
 }
+#ifdef MEMORY_DEBUG
+#include <mcheck.h>
+static PyObject* _py_mtrace(PyObject* self, PyObject* args)
+{
+	int startStop;
+	if(!PyArg_ParseTuple(args,"i:mtrace",&startStop)) return NULL;
+	if (startStop) {
+		mtrace();
+		}
+	else {
+		muntrace();
+		}
+	Py_INCREF(Py_None);
+	return Py_None;
+}
+#endif
 
 #define HEADER_SIZE		512
 
@@ -1958,7 +2033,7 @@ static int pict_putRow(BYTE_STREAM* fd, int row, int cols, pixel* rowpixels, cha
 		}
 	if (count > 0) *p++ = counttochar(count);
 
-	packcols = p - packed;		/* how many did we write? */
+	packcols = (int)(p - packed);		/* how many did we write? */
 	if (cols > 250){
 		pict_putShort(fd, packcols);
 		oc = packcols + 2;
@@ -1980,7 +2055,9 @@ static int pict_putRow(BYTE_STREAM* fd, int row, int cols, pixel* rowpixels, cha
 static PyObject* pil2pict(PyObject* self, PyObject* args)
 {
 	PyObject *result;
-	int		rows, cols, colors, i, row, oc, len, npixels, tc=-1;
+	Py_ssize_t	npixels, colors;
+	int		rows, cols, i, row, oc, tc=-1;
+	size_t	len;
 	char	*packed;
 	long	lpos;
 	pixel	*palette, *pixels;
@@ -2047,7 +2124,7 @@ static PyObject* pil2pict(PyObject* self, PyObject* args)
 	pict_putLong(obs, 0L);	/* pmReserved */
 	pict_putLong(obs, 0L);	/* ctSeed */
 	pict_putShort(obs, 0);	/* ctFlags */
-	pict_putShort(obs, colors-1);	/* ctSize */
+	pict_putShort(obs, (int)(colors-1));	/* ctSize */
 
 	/*Write out the colormap*/
 	for (i = 0; i < colors; i++){
@@ -2072,7 +2149,7 @@ static PyObject* pil2pict(PyObject* self, PyObject* args)
 	pict_putShort(obs, PICT_EndOfPicture);
 
 	len = obs->p-obs->buf;
-	lpos = (obs->p-obs->buf) - HEADER_SIZE;
+	lpos = (int)(obs->p-obs->buf) - HEADER_SIZE;
 	obs->p = obs->buf + HEADER_SIZE;
 	pict_putShort(obs, (short)(lpos & 0xffff));
 	result = PyBytes_FromStringAndSize((const char *)obs->buf,len);
@@ -2082,7 +2159,7 @@ static PyObject* pil2pict(PyObject* self, PyObject* args)
 
 
 static struct PyMethodDef _methods[] = {
-	{"gstate", (PyCFunction)gstate, METH_VARARGS|METH_KEYWORDS, "gstate(width,height[,depth=3][,bg=0xffffff]) create an initialised graphics state"},
+	{"gstate", (PyCFunction)gstate, METH_VARARGS|METH_KEYWORDS, "gstate(width,height[,depth=3][,bg=0xffffff]) create an\n        initialised graphics state"},
 	{"makeT1Font", (PyCFunction)makeT1Font, METH_VARARGS|METH_KEYWORDS, "makeT1Font(fontName,pfbPath,names)"},
 	{"delCache", (PyCFunction)delCache, METH_VARARGS, "delCache()"},
 	{"pil2pict", (PyCFunction)pil2pict, METH_VARARGS, "pil2pict(cols,rows,datastr,palette) return PICT version of im as bytes"},
@@ -2090,6 +2167,9 @@ static struct PyMethodDef _methods[] = {
     {"ft_get_face", (PyCFunction)ft_get_face, METH_VARARGS|METH_KEYWORDS,"ft_get_face(fontName) --> ft_face instance"},
 	{"parse_utf8", (PyCFunction)parse_utf8, METH_VARARGS, "parse_utf8(utf8_string) return UCS list"},
 #endif /*ifdef	RENDERPM_FT*/
+#ifdef MEMORY_DEBUG
+    {"mtrace", (PyCFunction)_py_mtrace, METH_VARARGS|METH_KEYWORDS,"mtrace(startStop) start or stop malloc tracing"},
+#endif /*MEMORY_DEBUG*/
 	{NULL,	NULL}			/*sentinel*/
 	};
 
@@ -2139,7 +2219,7 @@ void init_renderPM(void)
 	if(!obj)goto err;
 	PyModule_AddObject(m, "_version", obj);
 
-	obj = PyUnicode_FromString(LIBART_VERSION);
+	obj = PyUnicode_FromString(STRINGIFY(LIBART_VERSION));
 	if(!obj)goto err;
 	PyModule_AddObject(m, "_libart_version", obj);
 
